#  Copyright 2022, Olivier 'reivilibre'.
#
#  This file is part of Scone.
#
#  Scone is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Scone is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Scone.  If not, see <https://www.gnu.org/licenses/>.
import logging
from asyncio import Lock
from typing import Any, Dict, List, Optional, Tuple, Union

import attr
from HurricaneDNS import HurricaneDNS

from scone.head.head import Head
from scone.head.kitchen import Kitchen, Preparation
from scone.head.recipe import HeadRecipe, RecipeContext
from scone.head.utils import check_type

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class DnsRecord:
    subdomain: str
    kind: str
    value: str
    ttl: Optional[str]
    priority: Optional[str]


def parse_records(given_raw: Union[Any, Dict[str, Dict[str, str]]]) -> List[DnsRecord]:
    given = check_type(given_raw, dict, "records")

    the_list: List[DnsRecord] = []

    for key, attributes in given.items():
        # keys: "xyz A"
        # values: dicts, with keys:
        # - v: 1.2.3.4
        # - ttl: 86400
        # - priority: 50 (for MXes)

        pieces = key.split(" ")
        if len(pieces) > 2:
            raise ValueError(
                f"Key {key} should be space-separable with 2 or less pieces."
            )

        if len(pieces) == 2:
            subdomain, kind = pieces
        else:
            assert len(pieces) == 1
            (kind,) = pieces
            subdomain = ""

        ttl_raw = attributes.get("ttl")
        prio_raw = attributes.get("priority")

        record_value = attributes.get("v", attributes.get("value", None))

        if record_value is None:
            raise ValueError("No record value")

        the_list.append(
            DnsRecord(
                subdomain=subdomain,
                kind=kind,
                value=record_value,
                ttl=None if ttl_raw is None else str(ttl_raw),
                priority=None if prio_raw is None else str(prio_raw),
            )
        )

    return the_list


@attr.s(auto_attribs=True)
class HeRecord:
    id: str
    status: Optional[str]
    host: str
    type: str
    ttl: str
    # MX Priority
    mx: str
    value: str
    extended: str


@attr.s(auto_attribs=True)
class HeDomain:
    domain: str
    id: str
    type: str
    records: Optional[List[HeRecord]]


@attr.s(auto_attribs=True)
class HurricaneElectricCache:
    client: HurricaneDNS
    lock: Lock
    domains: Dict[str, HeDomain]


# Tuple from (id(head), user, password) → HE DNS cache)
# If Scone is ever long-living, this could leak, but it doesn't so it won't matter
HE_CLIENT_CACHE: Dict[Tuple[int, str, str], HurricaneElectricCache] = {}
CLIENT_CACHE_LOCK: Lock = Lock()


class HurricaneElectricDns(HeadRecipe):
    _NAME = "dns-hurricane"

    def __init__(self, recipe_context: RecipeContext, args: dict, head: Head):
        super().__init__(recipe_context, args, head)

        self.username = check_type(args.get("username"), str)
        self.password = check_type(args.get("password"), str)
        self.domain = check_type(args.get("domain"), str)

        self.records = parse_records(args.get("records"))

    def prepare(self, preparation: Preparation, head: Head) -> None:
        super().prepare(preparation, head)

    async def _get_client(self, head: Head) -> HurricaneElectricCache:
        async with CLIENT_CACHE_LOCK:
            cache_key = id(head), self.username, self.password
            if cache_key in HE_CLIENT_CACHE:
                # Happy days
                return HE_CLIENT_CACHE[cache_key]

            # TODO(performance): this takes about 3 sec; move it to an executor thread
            client = HurricaneDNS(self.username, self.password)

            domains = {}
            for domain in client.list_domains():
                dom = HeDomain(**domain)
                domains[dom.domain] = dom

            entry = HurricaneElectricCache(client, Lock(), domains)
            HE_CLIENT_CACHE[cache_key] = entry

            return entry

    async def _get_records(
        self, cached: HurricaneElectricCache, domain: HeDomain
    ) -> List[HeRecord]:
        async with cached.lock:
            if domain.records is not None:
                return domain.records

            domain.records = [
                HeRecord(**row) for row in cached.client.list_records(domain.domain)
            ]
            return domain.records

    async def cook(self, kitchen: Kitchen) -> None:
        # TODO(correctness): can't handle multiple DNS records
        #   with same (type, subdomain)
        kitchen.get_dependency_tracker().ignore()
        cached = await self._get_client(kitchen.head)

        records = await self._get_records(cached, cached.domains[self.domain])

        records_cache: Dict[Tuple[str, str], HeRecord] = {}

        for record in records:
            dotted_subdomain_suffix = f".{self.domain}"
            if record.host == self.domain:
                subdomain = ""
            elif record.host.endswith(dotted_subdomain_suffix):
                subdomain = record.host[: -len(dotted_subdomain_suffix)]
            else:
                raise ValueError(f"Can't figure out subdomain for {record.host}")

            records_cache[(subdomain, record.type)] = record

        logger.debug("Present records: %r", records_cache.keys())

        for wanted_record in self.records:
            wr_key = (wanted_record.subdomain, wanted_record.kind)
            logger.debug("Want %r: %r", wr_key, wanted_record)
            existing_record = records_cache.get(wr_key)
            if existing_record is not None:
                # TODO(correctness): amend as needed
                logger.debug("Found existing %r", existing_record)
            else:
                logger.debug("Will need to create new one")
                async with cached.lock:
                    cached.client.add_record(
                        self.domain,
                        wanted_record.subdomain,
                        wanted_record.kind,
                        wanted_record.value,
                        wanted_record.priority or None,
                        wanted_record.ttl or 86400,
                    )
