#  Copyright 2020, Olivier 'reivilibre'.
#
#  This file is part of Scone.
#
#  Scone is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Scone is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Scone.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path
from typing import List, Optional

from scone.common.modeutils import DEFAULT_MODE_FILE
from scone.default.steps.filesystem_steps import (
    template_jinja2_builtin,
    write_sous_file,
)
from scone.default.steps.systemd_steps import (
    cook_systemd_daemon_reload,
    cook_systemd_enable,
    cook_systemd_start,
    cook_systemd_stop,
)
from scone.head.head import Head
from scone.head.kitchen import Kitchen, Preparation
from scone.head.recipe import Recipe, RecipeContext
from scone.head.utils import check_type, check_type_opt


class SystemdUnit(Recipe):
    """
    System unit.
    TODO(performance): make it collapsible in a way so that it can daemon-reload
        only once in most situations.

    TODO(performance): deduplication.
    """

    _NAME = "systemd"

    def __init__(self, recipe_context: RecipeContext, args: dict, head):
        super().__init__(recipe_context, args, head)

        unit = check_type(args.get("unit"), str)
        self.unit_name = unit if "." in unit else unit + ".service"
        already_installed = check_type(args.get("already_installed", False), bool)
        self.at = check_type_opt(args.get("at", None), str)

        if not (already_installed or self.at):
            # already_installed is for when the unit already exists on the system
            # and is not created by scone.
            raise ValueError(
                "Must supply either already_installed = true or "
                f"at = /path/to/{self.unit_name}"
            )

        self.enabled = check_type_opt(args.get("enabled"), bool)
        self.restart_on = check_type_opt(args.get("restart_on"), list)
        self.started = check_type_opt(args.get("started"), bool)

    def prepare(self, preparation: Preparation, head: Head) -> None:
        super().prepare(preparation, head)
        if self.at:
            # TODO(potential future): preparation.provides("systemd-unit",
            #  self.unit_name)
            preparation.needs("file", self.at)

    async def cook(self, kitchen: Kitchen) -> None:
        if self.enabled is not None or self.started is not None:
            await cook_systemd_daemon_reload(kitchen)

        if self.enabled is not None:
            await cook_systemd_enable(kitchen, self.enabled, self.unit_name)

        if self.started is not None:
            if self.started:
                await cook_systemd_start(kitchen, self.unit_name)
            else:
                await cook_systemd_stop(kitchen, self.unit_name)


class SystemdTimer(Recipe):
    """
    Shorthand for creating a Systemd Service and associated Timer.

    Two vague kinds of timer:
    - monotonic timers
        e.g. OnBootSec, OnUnitActiveSec
        They stop if the computer is temporarily suspended or shut down.

        Not supported at present.
    - calendar timers
        specified by OnCalendar, in the format:
            `DayOfWeek Year-Month-Day Hour:Minute:Second`
        where you can use * for ALL, comma (,) for multiple and
        double-dot (..) for ranges.
    """

    _NAME = "systemd-timer"

    def __init__(self, recipe_context: RecipeContext, args: dict, head):
        super().__init__(recipe_context, args, head)

        self.unit_name = check_type(args.get("unit"), str)
        assert "." not in self.unit_name and "/" not in self.unit_name
        self.enabled = check_type(args.get("enabled", True), bool)

        self.working_directory = check_type(args.get("cd"), str)
        self.command = check_type(args.get("command"), str)
        self.env = check_type_opt(args.get("environment"), dict)

        self.user = check_type(args.get("user", recipe_context.user), str)
        self.group = check_type(args.get("group", self.user), str)

        self.description = check_type_opt(args.get("description"), str)

        self.calendar = args.get("calendar")
        if not (
            self.calendar is None
            or isinstance(self.calendar, str)
            or isinstance(self.calendar, list)
        ):
            raise TypeError(
                "`calendar` should either be omitted, a string or a list of strings."
            )
        self.persistent = check_type_opt(args.get("persistent"), bool)
        if self.calendar is None and self.persistent is not None:
            raise TypeError(
                "`persistent` may not be specified if `calendar` is not specified."
            )

        if self.calendar is None:
            raise TypeError(
                "Currently, only calendar/realtime timers are supported. "
                "You must specify `calendar`..."
            )

    def prepare(self, preparation: Preparation, head: Head) -> None:
        super().prepare(preparation, head)
        preparation.needs("os-user", self.user)
        preparation.needs("os-group", self.group)
        preparation.needs("directory", self.working_directory)

    async def cook(self, kitchen: Kitchen) -> None:
        timer_template_path = Path(__file__).parent.parent.joinpath(
            "templates", "systemd-timer.timer.j2"
        )
        service_template_path = Path(__file__).parent.parent.joinpath(
            "templates", "systemd-timer.service.j2"
        )

        if isinstance(self.calendar, str):
            calendars: Optional[List[str]] = [self.calendar]
        elif isinstance(self.calendar, list):
            calendars = self.calendar
        else:
            calendars = None

        timer_unit = template_jinja2_builtin(
            timer_template_path,
            calendars=calendars,
            persistent=self.persistent,
            description=self.description
            or f"{self.unit_name} (scone-generated timer timer)",
        )

        service_unit = template_jinja2_builtin(
            service_template_path,
            cd=self.working_directory,
            command=self.command,
            environment=self.env or {},
            user=self.user,
            group=self.group,
            description=self.description
            or f"{self.unit_name} (scone-generated timer service)",
        )

        await write_sous_file(
            kitchen,
            f"/etc/systemd/system/{self.unit_name}.timer",
            DEFAULT_MODE_FILE,
            timer_unit,
        )
        await write_sous_file(
            kitchen,
            f"/etc/systemd/system/{self.unit_name}.service",
            DEFAULT_MODE_FILE,
            service_unit,
        )

        if self.enabled is not None:
            await cook_systemd_daemon_reload(kitchen)
            await cook_systemd_enable(kitchen, self.enabled, self.unit_name + ".timer")
            await cook_systemd_start(kitchen, self.unit_name + ".timer")
