#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "scone"
DESCRIPTION = "Simple CONfiguration Engine"
URL = "https://librepush.net/project/scone"
EMAIL = "rei@librepush.net"
AUTHOR = "Olivier 'reivilibre'"
REQUIRES_PYTHON = ">=3.7.0"
VERSION = "0.2.0"

# What packages are required for this module to be executed?
REQUIRED = [
    "cbor2~=5.4.0",
    "setuptools~=49.1.2",
    "toml~=0.10.2",
    "attrs~=21.2.0",
    "cattrs>=1.7.1",
    "canonicaljson>=1.4.0",
    "immutabledict>=2.0.0",
]


EX_SOUS_BASE = []
EX_SOUS_PG = ["asyncpg"]
EX_SOUS_MYSQL = ["mysql-connector-python"]
EX_SOUS_DOCKER = ["docker"]

EX_SOUS_ALL = EX_SOUS_BASE + EX_SOUS_PG + EX_SOUS_MYSQL + EX_SOUS_DOCKER

EX_DEV_MYPY = [
    "types-toml",
    "types-requests"
]
EX_HEAD = [
    "SecretStorage~=3.1.2",
    "asyncssh[libnacl]~=2.2.1",
    "toposort~=1.5",
    "pynacl",
    "aiosqlite~=0.15.0",
    "requests",
    "Jinja2",
    "typeguard",
    "textx",
]
EX_HEAD_HEDNS = [
    "hurricanedns~=1.0.2",
]
EX_HEAD_ALL = EX_HEAD + EX_HEAD_HEDNS

EX_DEV = EX_SOUS_ALL + EX_HEAD + EX_DEV_MYPY

# What packages are optional?
EXTRAS = {
    "head-core": EX_HEAD,
    "head": EX_HEAD_ALL,
    "head-hedns": EX_HEAD_HEDNS,
    "sous": EX_SOUS_ALL,
    "sous-core": EX_SOUS_BASE,
    "sous-pg": EX_SOUS_PG,
    "sous-mysql": EX_SOUS_MYSQL,
    "sous-docker": EX_SOUS_DOCKER,
    "mypy": EX_DEV_MYPY,
    "dev": EX_DEV
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    entry_points={
        "console_scripts": [
            "scone=scone.head.cli:cli",
            "scone-freezer=scone.head.cli.freezer:cli",
            "michelin=scone.head.cli.michelin:cli",
        ],
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    # TODO license='GPL3',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand,},
)
